# coding=utf-8
import click
from acfunsdk import Acer
from acsaver import AcSaver, live_recorder

__author__ = 'dolacmeo'


@click.command()
@click.argument('cmd_name')
@click.option("-url", type=str)
@click.option("-root", type=str)
@click.option("--args", nargs=3)
def cli(cmd_name, url, root, args):
    if cmd_name == 'live_recorder':
        acer = Acer()
        live_obj = acer.get(args[0]).live
        return live_recorder(live_obj, *args[1:])
    elif cmd_name == "download":
        obj = AcSaver(Acer(), root).get(url)
        return obj.save_all()
    return True


if __name__ == '__main__':
    cli()



# coding=utf-8
from .utils import os, SaverBase, SaverData, downloader, tans_uub2html

__author__ = 'dolacmeo'


class MomentSaver(SaverBase):

    def __init__(self, acer, ac_obj):
        self.acer = acer
        self.ac_obj = ac_obj
        super().__init__(acer, ac_obj)

    @property
    def html_content(self):
        html_str, img_task = tans_uub2html(self.ac_obj.raw_data.get('text'), self._save_path)
        if len(img_task) > 0:
            downloader(self.acer.client, img_task, display=True)
        return html_str.replace(r"\"", '"').replace('emot/big', 'emot/small')

    @property
    def images(self):
        imgs = self.ac_obj.raw_data.get("imgs", [])
        if len(imgs) == 0:
            return imgs
        img_task = list()
        for i in imgs:
            filename = i['originUrl'].split("/")[-1]
            s = os.path.join(self._data_path, filename)
            img_task.append((i['originUrl'], s))
        res = downloader(self.acer.client, img_task)
        for i, x in enumerate(imgs):
            is_ok = res.get(x['originUrl'], False)
            if is_ok is True:
                filename = x['originUrl'].split("/")[-1]
                imgs[i]['originUrl'] = f"data/{filename}"
        return imgs

    def _gen_html(self):
        page_html = self.page_template.render(dict(saver=self, rtype_names=SaverData.rtype_name_map))
        html_path = os.path.join(self._save_path, f"{self.rid}.html")
        with open(html_path, 'wb') as html_file:
            html_file.write(page_html.encode())
        return os.path.isfile(html_path)

    def _save_image(self):
        img_task = [
            (self.ac_obj.qrcode, os.path.join(self._data_path, 'share_qrcode.png')),
            (self.ac_obj.mobile_qrcode, os.path.join(self._data_path, 'mobile_qrcode.png'))
        ]
        self.tasks.extend(img_task)
        t = downloader(self.acer.client, self.tasks, display=True)
        self.tasks = list()
        return t

    def save_all(self):
        self._save_raw()
        self._save_image()
        self._gen_html()
        self._save_comment()
        self.update_js_data()



# coding=utf-8
from .utils import os, Bs, SaverBase, downloader

__author__ = 'dolacmeo'


class ArticleSaver(SaverBase):

    def __init__(self, acer, ac_obj):
        self.acer = acer
        self.ac_obj = ac_obj
        super().__init__(acer, ac_obj)

    def _gen_html(self) -> bool:
        page_html = self.page_template.render(dict(saver=self))
        html_obj = Bs(page_html, 'lxml')
        img_tasks = list()
        for img in html_obj.select('.article-content img'):
            if img.attrs['src'].startswith('http') or img.attrs['src'].startswith('//'):
                img_name = img.attrs['src'].split('/')[-1]
                saved_path = os.path.join(self._img_path, img_name)
                img_tasks.append((img.attrs['src'], saved_path))
                img.attrs['alt'] = img.attrs['src']
                img.attrs['class'] = "lazy"
                img.attrs['src'] = "../../assets/img/logo-gray.png"
                img.attrs['data-src'] = f"./img/{img_name}"
        if len(img_tasks) > 0:
            downloader(self.acer.client, img_tasks, display=True)
        html_path = os.path.join(self._save_path, f"{self.rid}.html")
        with open(html_path, 'wb') as html_file:
            html_file.write(html_obj.prettify().encode())
        return os.path.isfile(html_path)

    def save_all(self):
        self._save_raw()
        self._save_image()
        self._gen_html()
        self._save_comment()
        self.update_js_data()



# coding=utf-8
from .utils import *

__author__ = 'dolacmeo'


class SaverBase:
    ac_obj = None
    keyname = None
    _save_root = None
    _save_path = None
    tasks = list()

    def __init__(self, acer, ac_obj):
        self.acer = acer
        self.ac_obj = ac_obj
        self.templates = saver_template()
        self.loading()

    @property
    def _objname(self):
        return self.__class__.__name__

    @property
    def _data_path(self):
        p = os.path.join(self._save_path, 'data')
        os.makedirs(p, exist_ok=True)
        return p

    @property
    def _img_path(self):
        p = os.path.join(self._save_path, 'img')
        os.makedirs(p, exist_ok=True)
        return p

    @property
    def rtype(self):
        return self.ac_obj.resource_type

    @property
    def rid(self):
        return self.ac_obj.resource_id

    @property
    def page_data(self):
        data = {"rType": self.rtype, "rId": self.rid, "objName": self._objname, "keyName": self.keyname}
        return json.dumps(data, separators=(',', ':'))

    @property
    def page_template(self):
        return self.templates.get_template(f"{self.keyname}.html")

    @property
    def comment_count(self):
        if self.rtype in [2, 3, 10]:
            return self.ac_obj.raw_data.get("commentCount", 0)
        elif self.rtype in [1]:
            return self.ac_obj.raw_data.get("data", {}).get("commentCount", 0)
        return 0

    def loading(self):
        assert self.ac_obj.__class__.__name__ in SaverData.ac_name_map.keys()
        self.keyname = SaverData.ac_name_map[self.ac_obj.__class__.__name__]
        self._save_root = self.acer.config.get("SaverRootPath", os.getcwd())
        self._save_path = os.path.join(self._save_root, self.keyname, str(self.rid))
        os.makedirs(self._save_path, exist_ok=True)
        pass

    def _save_raw(self):
        url_name = f"{self.ac_obj.title}"
        if self.rtype == 10:
            url_name = f"@{self.ac_obj.raw_data['user']['name']}"
        url_saved = url_saver(self.ac_obj.referer, self._save_path, url_name)
        raw_saved = json_saver(self.ac_obj.raw_data, self._data_path, f"{self.ac_obj.resource_id}")
        json2js(os.path.join(self._data_path, f"{self.rid}.json"), f"LOADED.{self.keyname}['{self.rid}']")
        if self.rtype not in [1]:
            self._save_member([self.ac_obj.up().uid])
        return all([url_saved, raw_saved])

    def _save_image(self):
        img_task = [
            (self.ac_obj.cover, os.path.join(self._save_path, 'cover._')),
            (self.ac_obj.qrcode, os.path.join(self._data_path, 'share_qrcode.png'))
        ]
        if self.keyname in ['video']:
            img_task.append((self.ac_obj.mobile_qrcode, os.path.join(self._data_path, 'mobile_qrcode.png')))
        self.tasks.extend(img_task)
        t = downloader(self.acer.client, self.tasks, display=True)
        self.tasks = list()
        return t

    def _part_video_name(self, num: int):
        assert self.keyname in ['video', 'bangumi']
        ends = "" if num == 0 else f"_{num + 1}"
        if self.keyname == 'bangumi':
            this_episode = self.ac_obj.episode_data[num]
            item_id = this_episode['itemId']
            vid = this_episode['videoId']
            ends = "" if num == 0 else f"_36188_{item_id}"
        return f"{self.rid}{ends}"

    def _gen_html(self) -> bool:
        page_html = self.page_template.render(dict(saver=self))
        html_path = os.path.join(self._save_path, f"{self.rid}.html")
        with open(html_path, 'wb') as html_file:
            html_file.write(page_html.encode())
        return os.path.isfile(html_path)

    def _save_video(self, num: int = 0, quality: [int, str] = 0):
        this_video = self.ac_obj.video(num)
        m3u8_url = this_video.m3u8_url(quality, False)
        vname = self._part_video_name(num)
        json_saver(this_video.raw_data, self._data_path, f"{vname}.video.json")
        scenes_data = this_video.scenes
        if isinstance(scenes_data, dict):
            json_saver(scenes_data, self._data_path, f"{vname}.scenes.json")
            scenes_to_thumbnails(self._save_path, vname)
        else:
            ffmpeg_gen_thumbnails(self._save_path, vname)
        hotspot_data = this_video.hotspot
        if isinstance(hotspot_data, dict):
            json_saver(hotspot_data, self._data_path, f"{vname}.hotspot.json")
        save_path = os.path.join(self._save_path, f"{vname}.mp4")
        if os.path.isfile(save_path):
            return True
        download_ok = m3u8_downloader(m3u8_url[0], save_path)
        retry_count = 0
        while download_ok is False and retry_count < 10:
            u = m3u8_url[1][retry_count % len(m3u8_url[1])]
            retry_count += 1
            download_ok = m3u8_downloader(u, save_path)
        return True

    def _record_live(self, quality: [int, str] = -1):
        live = self.ac_obj.live
        if live is None:
            return False
        if quality in [-1, "-1"]:
            quality = len(live.representation) - 1
        cmd_with_progress = [
            "start", "cmd", "/q", "/k",
            f"chcp 65001 && mode con cols=49 lines=4 && title AcLive({live.uid}) &&",
            "python", SaverData.cmd_path, "live_recorder",
            "--args",
            f"{self.ac_obj.referer}",
            f"{self._save_path}",
            f"{quality}"
        ]
        subprocess.Popen(cmd_with_progress, shell=True)

    def _save_danmaku(self, num: int = 0, quality: [int, str] = "1080p"):
        this_video = self.ac_obj.video(num)
        vname = self._part_video_name(num)
        json_saver(this_video.danmaku.danmaku_data, self._data_path, f"{vname}.danmaku.json")
        json2js(os.path.join(self._data_path, f"{vname}.danmaku.json"), f"LOADED.danmaku['{vname}']")
        ass_req = danmaku2ass(self._save_path, vname, quality)
        player_req = danmaku2dplayer(self._save_path, vname)
        return all([ass_req, player_req])

    def _save_comment(self, update: bool = False, with_user_data: bool = False):
        local_comment_data, local_comment_floors = None, []
        comment_json_path = os.path.join(self._data_path, f"{self.rid}.comment.json")
        comment_json_path_saved = os.path.isfile(comment_json_path)
        if comment_json_path_saved:
            local_comment_data = json.load(open(comment_json_path, 'rb'))
            local_comment_floors = [x['floor'] for x in local_comment_data['rootComments']]
        if update is True or comment_json_path_saved is False:
            comment_obj = self.ac_obj.comment()
            if self.comment_count < 1000:
                comment_obj.get_all_comments()
            else:
                comment_obj.get_all_comments(3)
            if local_comment_data is not None:
                comment_data = local_comment_data
                comment_data['hotComments'] = comment_obj.hot_comments
                comment_data['rootComments'].extend(
                    [c for c in comment_obj.root_comments if c['floor'] not in local_comment_floors])
                comment_data['subCommentsMap'].update(comment_obj.sub_comments)
                comment_data['save_unix'] = time.time()
            else:
                comment_data = {
                    "hotComments": comment_obj.hot_comments,
                    "rootComments": comment_obj.root_comments,
                    "subCommentsMap": comment_obj.sub_comments,
                    "save_unix": time.time()
                }
            uids = list()
            for c in comment_data['rootComments']:
                if c['userId'] not in uids:
                    uids.append(c['userId'])
            for _, i in comment_data['subCommentsMap'].items():
                for j in i['subComments']:
                    if j['userId'] not in uids:
                        uids.append(j['userId'])
            json_saver(comment_data, self._data_path, f"{self.rid}.comment.json")
            if with_user_data is True:
                self._save_member(uids)
        img_task = tans_comment_uub2html(self._save_path)
        if len(img_task) > 0:
            os.makedirs(os.path.join(self._save_path, 'img'), exist_ok=True)
            downloader(self.acer.client, img_task, display=True)
        return True

    def _save_member(self, ids: list, force: bool = False):
        if len(ids) == 0:
            return []
        done = list()
        member_dir = os.path.join(self._save_root, 'member')
        saved = os.listdir(member_dir)
        ids = sorted(list(set(ids)))
        ids_with_ext = [f"{i}.json" for i in ids]
        ids = list(set(ids_with_ext).difference([x for x in saved if x.endswith('.json')]))
        ids = [y.split('.')[0] for y in ids]
        avatar_task = list()
        with Progress(disable=len(ids) <= 5) as pp:
            get_member = pp.add_task("save members", total=len(ids))
            for uid in ids:
                if all([f"{uid}.json" in saved, f"{uid}.js" in saved, f"{uid}_avatar" in saved]) is True \
                        and force is False:
                    pp.update(get_member, advance=1)
                    done.append(uid)
                    continue
                user_req = self.acer.client.get(AcSource.apis['userInfo'], params=dict(userId=uid))
                user_data = user_req.json()
                profile = user_data.get('profile')
                user_json = os.path.join(member_dir, f"{uid}.json")
                user_js = os.path.join(member_dir, f"{uid}.js")
                user_avatar = os.path.join(member_dir, f"{uid}_avatar")
                if all([os.path.isfile(user_json), os.path.isfile(user_js), os.path.isfile(user_avatar)]) is True \
                        and force is False:
                    pp.update(get_member, advance=1)
                    done.append(uid)
                    continue
                with open(user_json, 'w') as uid_file:
                    json.dump(profile, uid_file, separators=(',', ':'))
                user_json_saved = os.path.isfile(user_json)
                user_json_string = open(os.path.join(user_json), 'rb').read().decode()
                with open(user_js, 'wb') as js_file:
                    user_js_string = f"let user_{uid}={user_json_string};"
                    js_file.write(user_js_string.encode())
                user_js_saved = os.path.isfile(user_js)
                avatar = parse.urlparse(profile['headUrl'])
                if not avatar.path.endswith('defaultAvatar.jpg'):
                    avatar_url = f"{avatar.scheme}://{avatar.netloc}{avatar.path}"
                    avatar_task.append((avatar_url, user_avatar))
                if all([user_json_saved, user_js_saved]):
                    done.append(uid)
                pp.update(get_member, advance=1)
                time.sleep(0.1)
            pp.update(get_member, completed=len(ids))
            pp.stop()
        if len(avatar_task) > 0:
            downloader(self.acer.client, avatar_task, display=True)
        return done

    def update_js_data(self):
        return update_js_data(self._save_root)



# coding=utf-8
import os
import re
import math
import time
import json
import httpx
import shutil
import random
import zipfile
import filetype
import subprocess
from urllib import parse
from PIL import Image
from bs4 import BeautifulSoup as Bs
from rich.console import Console
from rich.text import Text
from rich.align import Align
from rich.panel import Panel
from rich.live import Live
from rich.progress import Progress
from jinja2 import PackageLoader, Environment
from acfunsdk.source import AcSource
from .ffmpeg_progress_yield import FfmpegProgress
from .source import SaverData

__author__ = 'dolacmeo'
__all__ = (
    "os",
    "time",
    "json",
    "parse",
    "zipfile",
    "subprocess",
    "Console",
    "Text",
    "Align",
    "Panel",
    "Live",
    "Progress",
    "Bs",
    "AcSource",
    "SaverData",
    "FfmpegProgress",
    "saver_template",
    "sizeof_fmt",
    "unix2datestr",
    "url_saver",
    "json_saver",
    "json2js",
    "downloader",
    "m3u8_downloader",
    "ffmpeg_gen_thumbnails",
    "scenes_to_thumbnails",
    "danmaku2dplayer",
    "danmaku2ass",
    "get_usable_ffmpeg",
    "tans_uub2html",
    "tans_comment_uub2html",
    "live_recorder",
    "update_js_data",
    "create_http_server_bat"
)


def clean_file_name(filename:str):
    invalid_chars='[\\\/:*?"<>|]'
    replace_char='-'
    return re.sub(invalid_chars,replace_char,filename)


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Yi{suffix}"


def unix2datestr(t: (int, float, None) = None, f: str = "%Y-%m-%d %H:%M:%S"):
    if t is None:
        return time.strftime(f, time.localtime(time.time()))
    t = int(t)
    n = int(math.log10(t)) + 1
    if n > 10:
        t = t // math.pow(10, n - 10)
    elif n < 10:
        t = t * math.pow(10, 10 - n)
    return time.strftime(f, time.localtime(t))


def url_saver(url: str, base_path: [os.PathLike, str], filename: str):
    file_path = os.path.join(base_path, clean_file_name(filename))
    if not file_path.endswith(".url"):
        file_path = f"{file_path}.url"
    raw_data = f"[InternetShortcut]\nURL={url}\n"
    with open(file_path, 'wb') as url_file:
        url_file.write(raw_data.encode())
    result = os.path.isfile(file_path)
    return result


def json2js(src_path: [os.PathLike, str], keyname: str, dest_path: [str, None] = None):
    data = json.load(open(src_path, 'r'))
    if "moment" in src_path:
        text = tans_uub2html(data['text'], os.path.dirname(os.path.dirname(src_path)))
        data['text'] = text[0].replace(r"\"", '"').replace('emot/big', 'emot/small')

    str_data = json.dumps(data, separators=(',', ':'))
    data_js = f"{keyname}={str_data};"
    if dest_path is None:
        if src_path.endswith(".json"):
            dest_path = src_path[:-2]
        else:
            dest_path = os.path.join(os.path.dirname(src_path), f"{keyname}.js")
    with open(dest_path, 'wb') as js:
        js.write(data_js.encode())
    return os.path.isfile(dest_path)


def json_saver(data: dict, base_path: [os.PathLike, str], filename: str):
    file_path = os.path.join(base_path, filename)
    if not file_path.endswith(".json"):
        file_path = f"{file_path}.json"
    json_string = json.dumps(data, separators=(',', ':'))
    with open(file_path, 'wb') as json_file:
        json_file.write(json_string.encode())
    result = os.path.isfile(file_path)
    return result


def ffmpeg_gen_thumbnails(base_path: [os.PathLike, str], filename: str):
    media_path = os.path.join(base_path, f"{filename}.mp4")
    thumbnails_path = os.path.join(base_path, "data", f"{filename}.thumbnails.png")
    if os.path.isfile(thumbnails_path) is True:
        return True
    ffprobe_params = [
        "ffprobe", "-v", "error", "-show_entries", "stream=width,height,duration",
        "-of", "default=noprint_wrappers=1", "-print_format", "json", media_path
    ]
    p = subprocess.Popen(ffprobe_params, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    media_info = json.loads(p.stdout.read()).get("streams", [None])[0]
    w, h, d = media_info['width'], media_info['height'], media_info['duration']
    scale = "160:90" if w > h else "90:160"
    ffmpeg_cmd = get_usable_ffmpeg()
    ffmpeg_params = [
        ffmpeg_cmd, '-y', '-i', media_path, '-vsync', 'vfr', '-vf',
        f'fps=1/{float(d) / 100},scale={scale},tile=100x1',
        '-qscale:v', '3', thumbnails_path,
        # "-v", "quiet"
    ]
    # print(f"{ffmpeg_params=}")
    p = subprocess.Popen(ffmpeg_params, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.stdout.read()
    return os.path.isfile(thumbnails_path)


def scenes_to_thumbnails(base_path: [os.PathLike, str], filename: str):
    scenes_data_path = os.path.join(base_path, "data", f"{filename}.scenes.json")
    if not os.path.isfile(scenes_data_path):
        return False
    sprite = json.load(open(scenes_data_path, 'r'))
    pos = sprite['pos']
    if len(pos) % 100 != 0:
        if (len(pos) % 100) <= 10:
            pos = pos[:-(len(pos) % 100)]
        else:
            return ffmpeg_gen_thumbnails(base_path, filename)
    img_task = list()
    for i, x in enumerate(sprite['sprite_images']):
        img_task.append([x, os.path.join(base_path, "data", f"{filename}.scenes.{i}.png")])
    downloader(httpx.Client(), img_task, os.path.join(base_path, "data"))
    scenes0 = Image.open(img_task[0][1])
    ww, hh = scenes0.width // 5, scenes0.height // 10
    thumbnails = Image.new('RGB', (ww * 100, hh))
    step = len(pos) // 100
    for i in range(100):
        x, y, w, h = list(map(int, pos[i * step][2].split(',')))
        img = Image.open(img_task[pos[i * step][3]][1])
        region = img.crop((x, y, x + w, y + h))
        thumbnails.paste(region, (ww * i, 0))
    thumbnails_path = os.path.join(base_path, "data", f"{filename}.thumbnails.png")
    thumbnails.save(thumbnails_path, "png")
    return os.path.isfile(thumbnails_path)


def danmaku2dplayer(folder_path: str, filenameId: str):
    # 检查路径
    assert os.path.isdir(folder_path) is True
    danmaku_data_path = os.path.join(folder_path, 'data', f"{filenameId}.danmaku.json")
    assert os.path.isfile(danmaku_data_path) is True
    danmaku_data = json.load(open(danmaku_data_path, 'rb'))
    dplayer_data = dict(code=0, data=list())
    for danmaku in danmaku_data:
        dplayer_data['data'].append([
            round(danmaku['position'] / 1000, 3),
            danmaku['mode'] - 1,
            danmaku['color'],
            f"{danmaku['userId']}",
            danmaku['body']
        ])
    return json_saver(dplayer_data, os.path.join(folder_path, 'data'), f"{filenameId}.dplayer.json")


def danmaku2ass(folder_path: str, filenameId: str, vq: str = "720p", fontsize: int = 40):
    """
    https://github.com/niuchaobo/acfun-helper/blob/master/src/fg/modules/danmaku.js
    基础代码复刻自acfun助手中弹幕相关处理
    关于解决原代码中的弹幕重叠问题：
        0. 原弹幕数据要按时间进行排序
        1. 记录每条弹幕通道最后截止位置
        2. 如果同期所有通道已满，则减少弹幕停留时间(加速通过)

    :param client: acer.client
    :param folder_path: source path
    :param filenameId: ac_num
    :param vq: VideoQuality
    :return: ass file path
    :param fontsize: num px
    """
    # 检查路径
    assert os.path.isdir(folder_path) is True
    folder_name = os.path.basename(folder_path)
    main_json_name = f"{filenameId}.json"
    if "_" in filenameId:
        main_json_name = f"{filenameId.split('_')[0]}.json"
    main_data_path = os.path.join(folder_path, 'data', main_json_name)
    danmaku_data_path = os.path.join(folder_path, 'data', f"{filenameId}.danmaku.json")
    video_data_path = os.path.join(folder_path, 'data', f"{filenameId}.video.json")
    assert all([os.path.isfile(main_data_path), os.path.isfile(danmaku_data_path)]) is True
    main_data = json.load(open(main_data_path, 'rb'))
    danmaku_data = json.load(open(danmaku_data_path, 'rb'))
    video_data = json.load(open(video_data_path, 'rb'))
    ks_data = json.loads(video_data['ksPlayJson'])
    quality_data = None
    for x in ks_data['adaptationSet'][0]['representation']:
        if x['qualityType'] == vq:
            quality_data = x
    if isinstance(quality_data, dict) is False:
        return None
    if len(danmaku_data) == 0:
        return None
    thisVideoInfo = AcSource.videoQualitiesRefer[vq]
    thisVideoWidth = quality_data['width']
    thisVideoHeight = quality_data['height']
    thisDuration = 10
    channelNum = math.floor(thisVideoWidth / fontsize)
    if "dougaId" in main_data:
        scriptInfo = "\n".join([
            "[Script Info]",
            f"; AcVid: {folder_name}",
            f"; StreamName: {main_data['title']}",
            f"Title: {folder_name} - {main_data['user']['name']} - {main_data['title']}",
            f"Original Script: {folder_name} - {main_data['user']['name']} - {main_data['title']}",
            "Script Updated By: acfunSDK转换",
            "ScriptType: v4.00+",
            "Collisions: Normal",
            f"PlayResX: {thisVideoWidth}",
            f"PlayResY: {thisVideoHeight}"
        ])
    elif "bangumiId" in main_data.get("data", {}):
        scriptInfo = "\n".join([
            "[Script Info]",
            f"; AcVid: {folder_name}",
            f"; StreamName: {main_data['data']['bangumiTitle']}",
            f"Title: {folder_name} - {main_data['data']['bangumiTitle']}",
            f"Original Script: {folder_name} - {main_data['data']['bangumiTitle']}",
            "Script Updated By: acfunSDK转换",
            "ScriptType: v4.00+",
            "Collisions: Normal",
            f"PlayResX: {thisVideoWidth}",
            f"PlayResY: {thisVideoHeight}"
        ])
    styles = "\n".join([
        "[V4+ Styles]",
        "Format: " + ", ".join([
            'Name', 'Fontname', 'Fontsize', 'PrimaryColour', 'SecondaryColour', 'OutlineColour',
            'BackColour', 'Bold', 'Italic', 'Underline', 'StrikeOut', 'ScaleX', 'ScaleY',
            'Spacing', 'Angle', 'BorderStyle', 'Outline', 'Shadow', 'Alignment', 'MarginL',
            'MarginR', 'MarginV', 'Encoding']),
        "Style: " + ",".join([
            'Danmu', 'Microsoft YaHei', f'{fontsize}',
            '&H00FFFFFF', '&H00FFFFFF', '&H00000000', '&H00000000',
            '0', '0', '0', '0', '100', '100', '0', '0', '1', '1',
            '0', '2', '20', '20', '2', '0'])
    ])
    events = "\n".join([
        "[Events]",
        "Format: " + ", ".join([
            'Layer', 'Start', 'End', 'Style', 'Name',
            'MarginL', 'MarginR', 'MarginV', 'Effect', 'Text\n'])
    ])
    assData = list()
    screenChannel = [None for i in range(channelNum)]

    def timeProc(second, offset=0):
        second = second + offset
        minute = math.floor(second / 60)
        hours = math.floor(second / 60 / 60)
        minute = minute - hours * 60
        second = second - hours * 60 * 60 - minute * 60
        sec = second + offset
        return f"{hours:0>2}:{minute:0>2}:{sec:0>2.2f}"

    def choice_channel(startT, endT):
        # 按新时间移除频道占位
        empty = []
        for i, thisEnd in enumerate(screenChannel):
            if i in [0, 1]:
                continue
            elif thisEnd is None:
                empty.append(i)
            elif startT > thisEnd:
                screenChannel[i] = None
        # 无空位时返回空
        if len(empty) == 0:
            return None
        # 随机选择空位，记录结束时间，返回结果
        used = random.choice(empty)
        screenChannel[used] = endT
        return used

    for danmaku in danmaku_data:
        # 略过高级弹幕
        if danmaku['danmakuType'] != 0:
            continue
        # 弹幕挂载时间（文本）（弹幕左边界 接触到 视频的右边界）
        startTime = danmaku['position'] / 1000
        # 弹幕的长度
        danmakuLen = len(danmaku['body']) * fontsize
        danmakuLen_total = danmakuLen + thisVideoWidth
        # 运动到出界的时间点
        toLeftTime = startTime + thisDuration + (danmakuLen_total / thisVideoWidth)
        # 寻找频道
        danmaku_channel = choice_channel(startTime, toLeftTime)
        if danmaku_channel is None:  # 频道全满，加速通过
            toLeftTime -= int(thisDuration / 2)
            danmaku_channel = random.randint(2, channelNum)
        channelHeight = danmaku_channel * fontsize
        # 所有点位
        x1 = danmakuLen_total
        y1 = channelHeight
        x2 = - danmakuLen
        y2 = channelHeight
        dialogue = [
            "Dialogue: 0", timeProc(startTime), timeProc(toLeftTime),
            "Danmu", f"{danmaku['userId']}", "20", "20", "2", "",
            "{\\move(" + f"{x1}", f"{y1}", f"{x2}", f"{y2})" + "}" + f"{danmaku['body']}"
        ]
        assData.append(",".join(dialogue))
    events += "\n".join(assData)
    result = "\n\n".join([scriptInfo, styles, events])
    ass_path = os.path.join(folder_path, f"{filenameId}.ass")
    with open(ass_path, 'w', encoding="utf_8_sig") as ass_file:
        ass_file.write(result)
    return os.path.isfile(ass_path)


def get_usable_ffmpeg(cmd: [str, None] = None):
    cmds = ['ffmpeg', os.path.join(os.getcwd(), 'ffmpeg.exe')]
    if cmd is not None and os.path.isfile(cmd):
        cmds = [cmd] + cmds
    for x in cmds:
        try:
            p = subprocess.Popen([x, '-version'], stdin=subprocess.DEVNULL,
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()
            vers = str(out, 'utf-8').split('\n')[0].split()
            assert vers[0] == 'ffmpeg' and vers[2][0] > '0'
            return x
        except FileNotFoundError:
            continue
    return None


def m3u8_downloader(m3u8_url: str, save_path: [str, os.PathLike, None] = None):
    # ffmpeg 下载
    filename = os.path.split(save_path)[-1]
    ffmpeg_cmd = get_usable_ffmpeg()
    ffmpeg_params = [
        ffmpeg_cmd, '-y', '-i', m3u8_url,
        '-c', 'copy', '-bsf:a', 'aac_adtstoasc',
        '--', save_path
    ]
    try:
        ff = FfmpegProgress(ffmpeg_params)
        with Progress() as pp:
            print(f"{m3u8_url=}")
            ff_download = pp.add_task(filename, total=100)
            for progress in ff.run_command_with_progress():
                if progress > 0:
                    pp.update(ff_download, completed=progress)
            pp.update(ff_download, completed=100)
            pp.stop()
    except RuntimeError as e:
        print(f"RuntimeError.ERROR: downloader run {ffmpeg_params=}")
        return False
    except TypeError as e:
        print(f"TypeError.ERROR: downloader run {ffmpeg_params=}")
        return False
    except KeyboardInterrupt as e:
        os.remove(save_path)
        return False
    return os.path.isfile(save_path)


def downloader(client, src_urls_with_filename: list,
               dest_dir: [os.PathLike, str, None] = None,
               display: [bool, None] = None,
               force: bool = False) -> dict:
    assert isinstance(client, httpx.Client)
    assert len(src_urls_with_filename) > 0 and len(src_urls_with_filename[0]) == 2
    dest_dir = os.getcwd() if dest_dir is None else dest_dir
    assert os.path.isdir(dest_dir)
    total = len(src_urls_with_filename)
    display = (total > 5) if display is None else display
    done_mark = dict()
    with Progress(disable=not display) as pp:
        for src_url, filename in src_urls_with_filename:
            ext_guess = filename is None
            if filename is None:
                filename = src_url.split("/")[-1] if filename is None else filename
                filename = os.path.join(dest_dir, filename)
            if force is False and os.path.isfile(filename):
                continue
            try:
                with client.stream("GET", src_url) as resp:
                    if resp.status_code // 100 != 2:
                        continue
                    total = int(resp.headers.get("Content-Length", 0))
                    total = None if total == 0 else total // 1024
                    download_task = pp.add_task(os.path.basename(filename), total=(total or 100))
                    with open(filename, 'wb') as save_file:
                        for chunk in resp.iter_bytes():
                            save_file.write(chunk)
                            if total is None:
                                pp.update(download_task, advance=1)
                            else:
                                pp.update(download_task, completed=resp.num_bytes_downloaded // 1024)
                        pp.update(download_task, completed=total or 100)
            except (httpx.ConnectError, httpx.ConnectTimeout, httpx.ReadTimeout):
                print("httpx.ConnectError:", src_url)
                if os.path.isfile(filename):
                    os.remove(filename)
                if src_url.startswith("https://raw.githubusercontent.com"):
                    for x in SaverData.github_booster:
                        new_url = src_url.replace("raw.githubusercontent.com", x)
                        print(f"Retry With Github Booster: {new_url}")
                        resp = downloader(client, [(new_url, filename)])
                        if all(list(resp.values())):
                            break
                continue
            except ValueError as e:
                print("ValueError:", f"{src_url=}")
                continue
            except KeyboardInterrupt:
                if os.path.isfile(filename):
                    os.remove(filename)
                raise KeyboardInterrupt
            finally:
                done_mark[src_url] = os.path.isfile(filename)
                if os.path.isfile(filename) is True and ext_guess is True:
                    if '.' not in filename:
                        kind = filetype.guess_extension(filename)
                        if kind is not None:
                            new_fpath = ".".join([filename, kind])
                            shutil.move(filename, new_fpath)
    return done_mark


def tans_uub2html(src: str, save_path: str) -> tuple:
    ubb_tag_basic = {
        r"\r\n": "<br>",
        "[b]": "<b>", "[/b]": "</b>",
        "[i]": "<i>", "[/i]": "</i>",
        "[u]": "<u>", "[/u]": "</u>",
        "[s]": "<s>", "[/s]": "</s>",
        "[/color]": r"</color>"
    }
    ubb_tag_rex = {
        "color": r"(\[color=(#[a-f0-9]{6})\])",
        "size": r"(\[size=(\d+px)\]([^\[]+)\[/size\])",
        "emot": r"(\[emot=acfun,(\S+?)\/])",
        "emot_old": r"(\[emot=([a-z0-9]+),(\S+?)\/])",
        "image": r"(\[img=\\u56fe\\u7247\]([^\[]*)\[\/img\])",
        "at": r"(\[at uid=(\d+)\](@[^\[]+)\[/at\])",
        "at_old": r"(\[at\]([^\[]+)\[/at\])",
        "resource": r"(\[resource id=(\d+) type=([1-5]) icon=[^\]]*\]([^\[]*)\[\/resource\])",
        "jump": r"(\[time duration=(\d+)\]([^\[]+)\[/time\])",
    }
    ubb_resource_url = {
        "1": AcSource.routes['bangumi'],
        "2": AcSource.routes['video'],
        "3": AcSource.routes['article'],
        "4": AcSource.routes['album'],
        "5": AcSource.routes['up']
    }
    ubb_resource_icon = {
        "1": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_video_pc.png\">',
        "2": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_pc_vid_18_3.png\">',
        "3": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_popup_article_pc.png\">',
        "4": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_heji_pc.png\">',
        "5": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_human_pc.png\">',
    }
    root_path = os.path.dirname(os.path.dirname(save_path))
    emot_map_path = os.path.join(root_path, 'assets', 'emot', 'emotion_map.json')
    emot_map = json.load(open(emot_map_path, 'r'))
    # 基础替换：换行,加粗,斜体,下划线,删除线,颜色结尾
    for ubb, tag in ubb_tag_basic.items():
        src = src.replace(ubb, tag)
    # 正则替换：颜色,表情,图片
    img_task = list()
    for n, rex_rule in ubb_tag_rex.items():
        for tag in re.compile(rex_rule).findall(src):
            if n == 'color':
                src = src.replace(tag[0], f'<font color=\\"{tag[1]}\\">')
            elif n == 'size':
                src = src.replace(tag[0], f'<span style=\\"font-size:{tag[1]}\\">{tag[2]}</span>')
            elif n == 'emot':
                if tag[1] in emot_map:
                    src = src.replace(tag[0], f'<img class=\\"ubb-emotion\\" src=\\"{emot_map[tag[1]]}\\">')
            elif n == 'emot_old':
                alias = ",".join(tag[1:])
                if alias in emot_map['saved']:
                    src = src.replace(tag[0], f'<img class=\\"ubb-emotion\\" src=\\"{emot_map["saved"][alias]}\\">')
                elif alias in emot_map['lost']:
                    pass
            elif n == 'image':
                img_name = tag[1].split('/')[-1]
                img_path = os.path.join(save_path, 'img', img_name)
                img_task.append((tag[1], img_path))
                src = src.replace(tag[0], f'<img class=\\"lazy\\" src=\\"../../assets/img/404img.png\\" data-src=\\"img/{img_name}\\" alt=\\"{tag[1]}\\">')
            elif n == 'at':
                src = src.replace(tag[0], f'<a class=\\"ubb-name\\" target=\\"_blank\\" href=\\"https://www.acfun.cn/u/{tag[1]}\\">{tag[2]}</a>')
            elif n == 'at_old':
                src = src.replace(tag[0], f'<a class=\\"ubb-name\\">@{tag[1]}</a>')
            elif n == 'jump':
                src = src.replace(tag[0], f'<a class=\\"quickJump\\" onclick=\\"SAVER.utils.quickJump({tag[1]})\\">{tag[2]}</a>')
            elif n == 'resource':
                resource_a = '<a class=\\"ubb-ac\\" data-aid=\\"{ac_num}\\" href=\\"{href}\\" target=\\"_blank\\">{title}</a>'
                src = src.replace(
                    tag[0], ubb_resource_icon[tag[2]] + resource_a.format(
                        ac_num=tag[1],
                        href=ubb_resource_url[tag[2]] + tag[1],
                        title=tag[3]
                    ))
    # for tag in re.compile(r"(\[([^\]]+)\])").findall(src):
    #     print(tag)
    return src, img_task


def tans_comment_uub2html(data_path):
    rid = os.path.basename(data_path)
    comment_json_path = os.path.join(data_path, 'data', f"{rid}.comment.json")
    comment_json_temp = os.path.join(data_path, 'data', f"{rid}.comment.temp")
    temp_ok = os.path.isfile(comment_json_temp)
    img_task = list()
    if temp_ok:
        comment_json_string = open(comment_json_temp, 'r').read()
    else:
        comment_json_string = open(comment_json_path, 'r').read()
        comment_json_string, img_task = tans_uub2html(comment_json_string, data_path)
        with open(comment_json_temp, 'w') as t:
            t.write(comment_json_string)
    comment_data = json.loads(comment_json_string)
    total_comment = len(comment_data['rootComments'])
    # 评论分块存储，每块100条；跟楼按每页划分。
    # 区块正向划分，预留已删除位置；区块顺序列表倒置；热评在最后。
    total_block = math.ceil(total_comment / 100)
    blocks = {}
    count = 0
    for X in comment_data['rootComments']:
        count += 1
        z = str(math.floor(X['floor'] / 100))
        if z not in blocks:
            blocks[z] = {
                "hotComments": [],
                "rootComments": [],
                "subCommentsMap": {},
                "save_unix": time.time(),
                "totalComment": total_comment
            }
        blocks[z]['rootComments'].append(X)
        cid = str(X['commentId'])
        if cid in comment_data['subCommentsMap']:
            blocks[z]['subCommentsMap'][cid] = comment_data['subCommentsMap'][cid]
    totals = 0
    for v in blocks.values():
        totals += len(v["rootComments"])
    if totals == 0:
        comment_block_js_path = os.path.join(data_path, 'data', f"{rid}.comment.1.js")
        with open(comment_block_js_path, 'wb') as js_file:
            comment_js = "LOADED.comment[1]={'hotComments':[],'rootComments':[],'subCommentsMap':{},};"
            js_file.write(comment_js.encode())
        return img_task
    blocks = [j for i, j in sorted(blocks.items(), reverse=True)]
    blocks[0]["hotComments"] = comment_data["hotComments"]
    for Y in blocks[0]["hotComments"]:
        cid = str(Y['commentId'])
        if cid not in blocks[0]['subCommentsMap'] and cid in comment_data['subCommentsMap']:
            blocks[0]['subCommentsMap'][cid] = comment_data['subCommentsMap'][cid]
    for i in range(len(blocks)):
        B = blocks[i]
        B.update({'page': i + 1, 'total': len(blocks)})
        B['rootComments'] = sorted(B['rootComments'], key=lambda x: x['floor'], reverse=True)
        comment_block_js_path = os.path.join(data_path, 'data', f"{rid}.comment.{i+1}.js")
        comment_block_js_string = json.dumps(B, separators=(',', ':'))
        with open(comment_block_js_path, 'wb') as js_file:
            comment_js = f"LOADED.comment[{i+1}]={comment_block_js_string};"
            js_file.write(comment_js.encode())
    if temp_ok:
        os.remove(comment_json_temp)
    return img_task


def saver_template(**filters):
    templates = Environment(loader=PackageLoader('acsaver', 'templates'))
    templates.filters['unix2datestr'] = unix2datestr
    templates.filters['math_ceil'] = math.ceil
    for name, func in filters.items():
        templates.filters[name] = func
    return templates


def live_recorder(live_obj, save_path: [os.PathLike, str], quality: [int, str] = 0):
    assert live_obj.is_open is True
    if isinstance(quality, str) and quality.isdigit():
        quality = int(quality)
    begin_time = live_obj.start_time.replace("-", "").replace(":", "").replace(" ", "")
    save_dir = os.path.join(save_path, begin_time)
    time_now = unix2datestr(f="%Y%m%d%H%M%S")
    save_mp4 = os.path.join(save_dir, f"{time_now}.mp4")
    os.makedirs(save_dir, exist_ok=True)
    adapt = live_obj.m3u8_url(quality, False)
    live_obs_stream = adapt['url']
    stream_split = parse.urlsplit(live_obs_stream)
    stream_key = parse.parse_qs(stream_split.query).get('auth_key', [])[0]
    live_obs_stream = f"{stream_split.scheme}://{stream_split.netloc}{stream_split.path}?auth_key={stream_key}"
    ffmpeg = get_usable_ffmpeg()
    if ffmpeg is None:
        print(f"ERROR: record need ffmpeg")
        return False
    cmd_with_progress = [
        ffmpeg,
        "-progress", "-", "-nostats",
        '-loglevel', '+repeat',
        "-i", f"{live_obs_stream}",
        "-c:v", "copy", "-c:a", "copy",
        f"{save_mp4}"
    ]
    p = subprocess.Popen(
        cmd_with_progress,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=False)
    begin_read = False
    tmp = dict()
    console = Console()

    def display_tui(data):
        filesize = int(data.get('total_size', 0))
        infos = f" 已录制 {data.get('out_time', '00:00:00.000000')}\r\n " \
                f" 比特率: {data.get('bitrate', '???')}  " \
                f" 大小: {sizeof_fmt(filesize): >6} "
        record_panel = Panel(Text(infos, justify='center'),
                             title=f"AcLive({live_obj.uid})@{time_now}.mp4",
                             border_style='red', width=50, style="black on white")
        return record_panel

    with Live(console=console) as live_console:
        while True:
            if p.stdout is None:
                continue
            stderr_line = p.stdout.readline().decode("utf-8", errors="replace").strip()
            if stderr_line == "" and p.poll() is not None:
                break
            if stderr_line == "Press [q] to stop, [?] for help":
                begin_read = True
                continue
            if begin_read is True:
                r = stderr_line.split('=')
                tmp.update({r[0]: r[1]})
                live_console.update(Align.center(display_tui(tmp)))

    return True


def update_js_data(save_root: [os.PathLike, str]):
    saver_data = dict()
    saver_data_path = os.path.join(save_root, "data.js")
    if os.path.isfile(saver_data_path):
        with open(saver_data_path, 'r') as js_string:
            saver_data = json.loads(js_string.read()[12:-1])  # let AcSaver=.....;
    least = saver_data.get('least', [])
    for fn in SaverData.folder_names:
        if fn in ['member', 'live']:
            continue
        old_dirs = saver_data.get(fn, [])
        fpath = os.path.join(save_root, fn)
        now_dirs = [i for i in os.listdir(fpath) if os.path.isdir(os.path.join(fpath, i))]
        news = list(set(now_dirs).difference(set(old_dirs)))
        nomore = list(set(old_dirs).difference(set(now_dirs)))
        final_dirs = list()
        for x in old_dirs:
            if x not in nomore:
                final_dirs.append(x)
            least_v = (fn, x)
            if x in nomore and least_v in least:
                least.remove(least_v)
        for y in news:
            if y not in final_dirs:
                final_dirs.append(y)
        final_dirs = [n for n in final_dirs if os.path.isfile(os.path.join(save_root, fn, n, 'data', f'{n}.js'))]
        saver_data[fn] = final_dirs
        for i in news:
            if [fn, i] not in news:
                least.append([fn, i])
    new = list()
    for u in least:
        if u not in new and u[1] in saver_data[u[0]]:
            new.append(u)
    saver_data['least'] = new
    with open(saver_data_path, 'wb') as js:
        saver_data_string = json.dumps(saver_data, separators=(',', ':'))
        data_js = f"let AcSaver={saver_data_string};"
        js.write(data_js.encode())
    return os.path.isfile(saver_data_path)


def create_http_server_bat(save_root: [os.PathLike, str]):
    bat_path = os.path.join(save_root, "CivetWeb.bat")
    if os.path.isfile(bat_path) is True:
        return True
    cmd = "@echo off\n" \
          "if exist CivetWeb64.exe (set \"exename=CivetWeb64.exe\" & goto run\n" \
          ") else if exist CivetWeb32.exe (set \"exename=CivetWeb32.exe\" & goto run\n" \
          ") else (echo \"需要CivetWeb64.exe或CivetWeb32.exe，赶紧去下载一个\" " \
          "& start https://github.com/civetweb/civetweb/releases & pause)\n" \
          ":run\nstart http://127.0.0.1:666/index.html\n" \
          "tasklist /fi \"ImageName eq %exename%\" /fo csv 2>NUL | find /I \"%exename%\">NUL\n" \
          "if \"%ERRORLEVEL%\"==\"1\" (start %exename% -listening_ports 666)\ngoto end"
    with open(bat_path, 'w') as bat:
        bat.write(cmd)
    return os.path.isfile(bat_path)




import re
import subprocess
from typing import Generator, List


def to_ms(string: str = None, precision: int = None, **kwargs) -> float:
    """
    Convert a string to milliseconds.
    You can either pass a string, or a set of keyword args ("hour", "min", "sec", "ms") to convert.
    If "precision" is set, the result is rounded to the number of decimals given.
    From: https://gist.github.com/Hellowlol/5f8545e999259b4371c91ac223409209
    """
    if string:
        hour = int(string[0:2])
        minute = int(string[3:5])
        sec = int(string[6:8])
        ms = int(string[10:11])
    else:
        hour = int(kwargs.get("hour", 0))
        minute = int(kwargs.get("min", 0))
        sec = int(kwargs.get("sec", 0))
        ms = int(kwargs.get("ms", 0))

    result = (hour * 60 * 60 * 1000) + (minute * 60 * 1000) + (sec * 1000) + ms
    if precision and isinstance(precision, int):
        return round(result, precision)
    return result


class FfmpegProgress:
    DUR_REGEX = re.compile(r"Duration: (?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<ms>\d{2})")
    TIME_REGEX = re.compile(r"out_time=(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<ms>\d{2})")

    def __init__(self, cmd: List[str], dry_run=False) -> None:
        """Initialize the FfmpegProgress class.

        Args:
            cmd (List[str]): A list of command line elements, e.g. ["ffmpeg", "-i", ...]
            dry_run (bool, optional): Only show what would be done. Defaults to False.
        """
        self.cmd = cmd
        self.dry_run = dry_run
        self.stderr = None

    def run_command_with_progress(self, popen_kwargs=None) -> Generator[int, None, None]:
        """
        Run an ffmpeg command, trying to capture the process output and calculate
        the duration / progress.
        Yields the progress in percent.

        Args:
            popen_kwargs (dict): A dict to specify extra arguments to the popen call, e.g. { creationflags: CREATE_NO_WINDOW }
        """
        if popen_kwargs is None:
            popen_kwargs = {}
        if self.dry_run:
            return

        total_dur = None

        cmd_with_progress = (
            [self.cmd[0]]
            + ["-progress", "-", "-nostats"]
            + self.cmd[1:]
        )

        stderr = []

        p = subprocess.Popen(
            cmd_with_progress,
            stdin=subprocess.PIPE,  # Apply stdin isolation by creating separate pipe.
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=False,
            **popen_kwargs
        )

        yield 0

        while True:
            if p.stdout is None:
                continue

            stderr_line = p.stdout.readline().decode("utf-8", errors="replace").strip()

            if stderr_line == "" and p.poll() is not None:
                break

            stderr.append(stderr_line.strip())

            self.stderr = "\n".join(stderr)

            total_dur_match = FfmpegProgress.DUR_REGEX.search(stderr_line)
            if total_dur is None and total_dur_match:
                total_dur = total_dur_match.groupdict()
                total_dur = to_ms(**total_dur)
                continue
            if total_dur:
                progress_time = FfmpegProgress.TIME_REGEX.search(stderr_line)
                if progress_time:
                    elapsed_time = to_ms(**progress_time.groupdict())
                    yield int(elapsed_time / total_dur * 100)

        if p.returncode != 0:
            raise RuntimeError(
                "Error running command {}: {}".format(self.cmd, str("\n".join(stderr)))
            )

        yield 100



# coding=utf-8
import os.path
from dataclasses import dataclass
from acfunsdk.source import AcSource

__author__ = 'dolacmeo'


@dataclass(frozen=True)
class SaverData:
    github_assets_map_url = "https://raw.githubusercontent.com/dolaCmeo/acsaver/assets/assets.map"
    github_assets_zip_url = "https://raw.githubusercontent.com/dolaCmeo/acsaver/assets/assets.zip"
    github_assets_base = "https://raw.githubusercontent.com/dolaCmeo/acsaver/assets/"
    github_booster = (
        "raw.fastgit.org",
        # "raw.githubusercontents.com",
        "raw.staticdn.net",
        "raw.iqiq.io"
    )
    folder_names = ['bangumi', 'article', 'video', 'live', 'member', 'moment']
    emot_alias = ['default', 'ac', 'ac2', 'ac3', 'dog', 'tsj', 'brd', 'ais', 'td', 'zuohe', 'blizzard']
    ac_name_map = {
        "AcArticle": "article",
        "AcMoment": "moment",
        "AcVideo": "video",
        "AcBangumi": "bangumi",
        "AcLiveUp": "live",
        "AcUp": "member"
    }
    ac_saver_map = {
        "AcBangumi": "BangumiSaver",
        "AcArticle": "ArticleSaver",
        "AcVideo": "VideoSaver",
        "AcUp": "MemberSaver",
        "AcMoment": "MomentSaver",
        "AcLiveUp": "LiveSaver",
    }
    rtype_name_map = {
        "1": "番剧",  # bangumi
        "2": "视频",  # video
        "3": "文章",  # article
        "4": "合辑",  # album
        "5": "用户",  # member
        "6": "评论",  # comment
        # "8": "私信",
        "10": "动态",  # moment
    }
    ubb_tag_basic = {
        r"\r\n": "<br>",
        "[b]": "<b>", "[/b]": "</b>",
        "[i]": "<i>", "[/i]": "</i>",
        "[u]": "<u>", "[/u]": "</u>",
        "[s]": "<s>", "[/s]": "</s>",
        "[/color]": r"</color>"
    }
    ubb_tag_rex = {
        "color": r"(\[color=(#[a-f0-9]{6})\])",
        "size": r"(\[size=(\d+px)\]([^\[]+)\[/size\])",
        "emot": r"(\[emot=acfun,(\S+?)\/])",
        "emot_old": r"(\[emot=([a-z0-9]+),(\S+?)\/])",
        "image": r"(\[img=\\u56fe\\u7247\]([^\[]*)\[\/img\])",
        "at": r"(\[at uid=(\d+)\](@[^\[]+)\[/at\])",
        "at_old": r"(\[at\]([^\[]+)\[/at\])",
        "resource": r"(\[resource id=(\d+) type=([1-5]) icon=[^\]]*\]([^\[]*)\[\/resource\])",
        "jump": r"(\[time duration=(\d+)\]([^\[]+)\[/time\])",
    }
    ubb_resource_url = {
        "1": AcSource.routes['bangumi'],
        "2": AcSource.routes['video'],
        "3": AcSource.routes['article'],
        "4": AcSource.routes['album'],
        "5": AcSource.routes['up']
    }
    ubb_resource_icon = {
        "1": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_video_pc.png\">',
        "2": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_pc_vid_18_3.png\">',
        "3": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_popup_article_pc.png\">',
        "4": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_heji_pc.png\">',
        "5": r'<img class=\"ubb-icon\" src=\"../../assets/img/icon_comment_human_pc.png\">',
    }
    cmd_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "__main__.py")



# coding=utf-8
from .bangumi import BangumiSaver
from .video import VideoSaver
from .article import ArticleSaver
from .moment import MomentSaver
from .live import LiveSaver
from .utils import os, json, zipfile, SaverData, downloader, \
    saver_template, update_js_data, create_http_server_bat, live_recorder

__author__ = 'dolacmeo'
__all__ = (
    "AcSaver",
    "SaverLocal"
)


class AcSaver:

    def __init__(self, acer, root_path: [os.PathLike, str, None] = None):
        self.acer = acer
        if root_path is None:
            root_path = os.getcwd()
        elif not os.path.exists(root_path):
            os.makedirs(root_path, exist_ok=True)
        self.acer.config.update({"SaverRootPath": root_path})
        self.local = SaverLocal(self.acer, root_path)
        self.loading()

    def loading(self):
        update_js_data(self.local.root_path)

    def BangumiSaver(self, ac_obj):
        return BangumiSaver(self.acer, ac_obj)

    def VideoSaver(self, ac_obj):
        return VideoSaver(self.acer, ac_obj)

    def ArticleSaver(self, ac_obj):
        return ArticleSaver(self.acer, ac_obj)

    def MomentSaver(self, ac_obj):
        return MomentSaver(self.acer, ac_obj)

    def LiveSaver(self, ac_obj):
        return LiveSaver(self.acer, ac_obj)

    def get_saver(self, ac_obj):
        s = SaverData.ac_saver_map.get(ac_obj.__class__.__name__)
        if s is None:
            return None
        return getattr(self, s)(ac_obj)

    def resource(self, rtype: int, rid: int):
        obj = self.acer.acfun.resource(rtype, rid)
        if obj is None:
            return None
        return self.get_saver(obj)

    def get(self, url_str: str):
        obj = self.acer.get(url_str)
        if obj is None:
            return None
        return self.get_saver(obj)


class SaverLocal:
    root_path = None
    __assets_check_times = 0

    def __init__(self, acer, root_path: [os.PathLike, str]):
        self.acer = acer
        self.root_path = root_path
        self.loading()

    def history(self):
        saver_data_path = os.path.join(self.root_path, "data.js")
        if os.path.isfile(saver_data_path) is False:
            return None
        with open(saver_data_path, 'r') as js_string:
            saver_data = json.loads(js_string.read()[12:-1])  # let AcSaver=.....;
        return saver_data

    def loading(self):
        self._folder_check()
        self._assets_check()
        self._page_check()
        create_http_server_bat(self.root_path)
        pass

    def _folder_check(self):
        # 必要目录检查
        done_mark = list()
        for x in SaverData.folder_names:
            f = os.path.join(self.root_path, x)
            os.makedirs(f, exist_ok=True)
            done_mark.append(os.path.isdir(f))
        return all(done_mark)

    def _assets_check(self):
        self.__assets_check_times += 1
        # 检查静态文件目录
        os.makedirs(os.path.join(self.root_path, "assets"), exist_ok=True)
        # 查找静态文件map
        assets_map_path = os.path.join(self.root_path, "assets.map")
        if os.path.isfile(assets_map_path) is False:
            # 如map不存在，认为静态文件缺失
            task = (SaverData.github_assets_map_url, assets_map_path)
            # 优先获取完整的map文件
            downloader(self.acer.client, [task], display=True)
            pass
        assert os.path.isfile(assets_map_path) is True
        # 读取map文件，进行本地文件校验
        assets_map_file = open(assets_map_path, 'rb')
        assets_map_data = assets_map_file.read().decode().split("\n")
        missing_assets = list()
        for x in assets_map_data:
            if not os.path.exists(os.path.join(self.root_path, x)):
                missing_assets.append(x)
        assets_map_file.close()
        # 如未通过校验，缺失少于10个
        if len(missing_assets) > 0:
            if len(missing_assets) <= 10:
                # 逐个将丢失文件重新下载
                tasks = list()
                for item in missing_assets:
                    url_end = item.replace('\\', '/')
                    tasks.append((f"{SaverData.github_assets_base}{url_end}",
                                  os.path.join(self.root_path, item)))
                downloader(self.acer.client, tasks)
            else:
                # 下载完整压缩包，覆盖解压处理
                assert_zip_path = os.path.join(self.root_path, "assets.zip")
                if not os.path.isfile(assert_zip_path):
                    downloader(self.acer.client, [(SaverData.github_assets_zip_url, assert_zip_path)], display=True)
                zip_file = zipfile.ZipFile(assert_zip_path)
                zip_file.extractall(self.root_path)
            # 循环调用检查，并计数，最大次数3次后报错
            if self.__assets_check_times > 3:
                raise FileNotFoundError("Assets Files Check MaxTried.")
            return self._assets_check()
        return True

    def _page_check(self):
        index_names = [
            ['index', '内容', 'https://www.acfun.cn/'],
            ['anime', '番剧', 'https://www.acfun.cn/bangumilist'],
            ['feed', '动态', 'https://www.acfun.cn/member/feeds'],
            # 直播
        ]
        for page, title, referer in index_names:
            index_html_path = os.path.join(self.root_path, f'{page}.html')
            if os.path.isfile(index_html_path) is False:
                templates = saver_template()
                index_html = templates.get_template(f'{page}.html').render(
                    main=dict(zip(["name", "title", "referer"], [page, title, referer])))
                with open(index_html_path, 'wb') as index_file:
                    index_file.write(index_html.encode())
        return True



# coding=utf-8
from .utils import os, SaverBase

__author__ = 'dolacmeo'


class VideoSaver(SaverBase):

    def __init__(self, acer, ac_obj):
        self.acer = acer
        self.ac_obj = ac_obj
        super().__init__(acer, ac_obj)

    @property
    def video_list(self):
        return self.ac_obj.video_list

    def save_all(self):
        self._save_raw()
        self._save_image()
        self._gen_html()
        for n in range(len(self.video_list)):
            self._save_video(n)
            self._save_danmaku(n)
        self._save_comment()
        self.update_js_data()



# coding=utf-8
from .utils import os, SaverBase, SaverData, url_saver, json_saver, json2js, downloader

__author__ = 'dolacmeo'


class LiveSaver(SaverBase):

    def __init__(self, acer, ac_obj):
        self.acer = acer
        self.ac_obj = ac_obj
        super().__init__(acer, ac_obj)

    @property
    def begin_time(self):
        if self.ac_obj.past_time > 0:
            return self.ac_obj.live.start_time.replace("-", "").replace(":", "").replace(" ", "")
        return None

    @property
    def save_dir(self):
        return os.path.join(self._save_path, self.begin_time)

    def live_raw_save(self):
        if self.ac_obj.past_time == -1:
            return False
        os.makedirs(self.save_dir, exist_ok=True)
        url_name = f"{self.ac_obj.live.raw_data.get('caption', '@'+self.ac_obj.username)}"
        url_saved = url_saver(self.ac_obj.referer, self.save_dir, url_name)
        raw_saved = json_saver(self.ac_obj.live.raw_data, self.save_dir, f"{self.begin_time}")
        json2js(os.path.join(self.save_dir, f"{self.begin_time}.json"),
                f"LOADED.live['{self.ac_obj.uid}_{self.begin_time}']")
        downloader(self.acer.client, [(self.ac_obj.cover, os.path.join(self.save_dir, 'cover._'))])
        return all([url_saved, raw_saved])

    def loading(self):
        assert self.ac_obj.__class__.__name__ in SaverData.ac_name_map.keys()
        self.keyname = SaverData.ac_name_map[self.ac_obj.__class__.__name__]
        self._save_root = self.acer.config.get("SaverRootPath", os.getcwd())
        self._save_path = os.path.join(self._save_root, self.keyname, str(self.ac_obj.uid))
        os.makedirs(self._save_path, exist_ok=True)

    def _save_raw(self):
        url_name = f"@{self.ac_obj.raw_data['user']['name']}"
        url_saved = url_saver(self.ac_obj.referer, self._save_path, url_name)
        self.tasks.extend([
            (self.ac_obj.qrcode, os.path.join(self._save_path, 'share_qrcode.png')),
            (self.ac_obj.mobile_qrcode, os.path.join(self._save_path, 'mobile_qrcode.png'))
        ])
        downloader(self.acer.client, self.tasks)
        self.tasks = list()
        self._save_member([self.ac_obj.uid])
        return url_saved

    def save_all(self):
        self._save_raw()
        self.live_raw_save()
        if self.ac_obj.past_time > 0:
            self._record_live()
        else:
            print(f"Live is CLOSED.")
        # 记录弹幕
        # 保存直播信息
        # self.update_js_data()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['acsaver', 'acsaver.utils']

package_data = \
{'': ['*'], 'acsaver': ['templates/*', 'templates/part/*']}

install_requires = \
['acfunsdk>=0.9.5,<0.10.0',
 'click>=8.1,<9.0',
 'filetype>=1.1,<2.0',
 'jinja2>=3.1,<4.0',
 'rich>=12.5,<13.0']

entry_points = \
{'console_scripts': ['acsaver = acsaver.__main__:cli']}

setup_kwargs = {
    'name': 'acsaver',
    'version': '0.1.3',
    'description': 'acfunsdk - AcSaver',
    'long_description': '# acfunsdk - AcSaver\n\n<br />\n\n<p align="center">\n<a href="https://github.com/dolaCmeo/acfunSDK">\n<img height="100" src="https://s3.dualstack.us-east-2.amazonaws.com/pythondotorg-assets/media/files/python-logo-only.svg" alt="">\n<img height="100" src="https://ali-imgs.acfun.cn/kos/nlav10360/static/common/widget/header/img/acfunlogo.11a9841251f31e1a3316.svg" alt="">\n</a>\n</p>\n\n<br />\n\nacfunsdk是 **非官方的 [AcFun弹幕视频网][acfun.cn]** Python库。\n\n> `acsaver` 是`acfunsdk`的附属组件，提供内容离线保存支持。\n\n- - -\n\n\n<details>\n<summary>依赖库</summary>\n\n>内置+修改: 位于 `libs` 文件夹内\n>\n>+ [`ffmpeg_progress_yield`](https://github.com/slhck/ffmpeg-progress-yield)\n\n**依赖: 包含在 `requirements.txt` 中**\n\n下载及html页面渲染:\n+ [`acfunsdk`](https://pypi.org/project/acfunsdk/)`>=0.9.5`\n+ [`filetype`](https://pypi.org/project/filetype/)`>=1.1`\n+ [`jinja2`](https://pypi.org/project/jinja2/)`>=3.1`\n+ [`rich`](https://pypi.org/project/rich/)`>=12.5`\n+ [`click`](https://pypi.org/project/click/)`>=8.1`\n\n</details>\n\n- - -\n\n## About Me\n\n[![ac彩娘-阿部高和](https://tx-free-imgs2.acfun.cn/kimg/bs2/zt-image-host/ChQwODliOGVhYzRjMTBmOGM0ZWY1ZRCIzNcv.gif)][dolacfun]\n[♂ 整点大香蕉🍌][acfunsdk_page]\n<img alt="AcFunCard" align="right" src="https://discovery.sunness.dev/39088">\n\n- - - \n\n[dolacfun]: https://www.acfun.cn/u/39088\n[acfunsdk_page]: https://www.acfun.cn/a/ac37416587\n\n[acfun.cn]: https://www.acfun.cn/\n[Issue]: https://github.com/dolaCmeo/acfunSDK/issues\n[python]: https://www.python.org/downloads/\n[venv]: https://docs.python.org/zh-cn/3.8/library/venv.html\n',
    'author': 'dolacmeo',
    'author_email': 'dolacmeo@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/acsaver/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
